/* 

                          Firewall Builder

                 Copyright (C) 2000 Vadim Kurland,  Vadim Zaliva

  Author:  Vadim Kurland     vadim@vk.crocodile.org
           Vadim Zaliva      lord@crocodile.org

  $Id: BackgroundOp.cc,v 1.10 2002/05/07 02:07:07 vkurland Exp $


  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <fwbuilder/libfwbuilder-config.h>

#include <stdlib.h>
#include <stdio.h>
#include <vector>
#include <iostream>
#ifdef __MINGW32__ //win32 pthread ditrib doesn't have config.h
# ifdef HAVE_CONFIG_H
#  undef HAVE_CONFIG_H
#  include <pthread.h>
#  define HAVE_CONFIG_H
# endif
#else
# include <pthread.h>
#endif
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <signal.h>

#include <glib.h>

#include <fwbuilder/BackgroundOp.hh>

using namespace std;
using namespace libfwbuilder;


BackgroundOp::BackgroundOp():stop_program(false),running(false),connected(true)
{
    error   = NULL;
    iamdead = new SyncFlag(false);
}

/*
 * we destroy logger in the background thread because run_impl may use
 * logger and we should be able to destroy object of this class even
 * while background thread is working. 
 *
 * iamdead should not be deleted either because it is used as a flag
 * so that background thread would know if this object has been deleted
 * while run_impl was running
 */
BackgroundOp::~BackgroundOp()
{
    iamdead->lock();
    iamdead->modify(true);
    iamdead->unlock();
}

bool BackgroundOp::isRunning()     { return running;  }
void BackgroundOp::setRunning()    { running=true;    }
void BackgroundOp::clearRunning()  { running=false;   }

bool BackgroundOp::isConnected()   { return connected; }
void BackgroundOp::disconnect()    { connected=false;  }



Logger* BackgroundOp::start_operation() throw(FWException)
{
    stop_program =false;
    running      =true;

    Logger *logger  = new QueueLogger();

    void **void_pair = new void*[3];
    void_pair[0] = this;
    void_pair[1] = logger;
    void_pair[2] = iamdead;

    pthread_t tid;
    pthread_create(&tid, NULL, background_thread, void_pair);
    pthread_detach(tid);

    return logger;
}

void BackgroundOp::stop_operation()
{
    error = new FWException("Interrupted by user");
    stop_program=true;
}

void *libfwbuilder::background_thread(void *args)
{
    void **void_pair=(void**)args;

    BackgroundOp *bop    = (BackgroundOp*)void_pair[0];
    Logger       *logger = (Logger *)     void_pair[1];
    SyncFlag     *isdead = (SyncFlag*)    void_pair[2];

    try 
    {
	bop->run_impl(logger);
    } catch (FWException &ex) 
    {
        isdead->lock();
        if (isdead->peek()) { 
            isdead->unlock(); 
            delete logger;
            delete isdead; 
            delete void_pair;
            return(NULL); 
        }
        *logger << "Exception: " << ex.toString() << '\n';
        bop->error=new FWException(ex);
        isdead->unlock();
    }

    isdead->lock();
    if (isdead->peek()) { 
        isdead->unlock(); 
        delete logger;
        delete isdead; 
        delete void_pair;
        return(NULL); 
    }

/* operation completed - clear "running" flag */
    bop->clearRunning();

    isdead->unlock();

/* wait till the other thread reads all the lines from logger. If
 * widget that was reading lines from this logger has been destroyed,
 * it should have called BackhroundOp::disconnect to release BackgroundOp
 * object
 */

    while (true) {
        
        isdead->lock();
        if (isdead->peek() ||  !bop->isConnected()) {
            isdead->unlock();
            break;
        }
        isdead->unlock();

#ifndef __MINGW32__
        sleep(1);
#else
        Sleep(1);
#endif
    }

    delete logger;
    delete void_pair;

    return(NULL);
}






