/* 

                          Firewall Builder

                 Copyright (C) 2000 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: TCPService.cc,v 1.11 2002/05/18 07:13:43 vkurland Exp $


  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include <fwbuilder/TCPService.hh>

#include <fwbuilder/XMLTools.hh>

using namespace libfwbuilder;
using namespace std;

const char *TCPService::TYPENAME={"TCPService"};

TCPService::TCPService()  
{
    setInt("src_range_start", 0);
    setInt("src_range_end",   0);
    setInt("dst_range_start", 0);
    setInt("dst_range_end",   0);

    setURGFlag(false);
    setACKFlag(false);
    setPSHFlag(false);
    setRSTFlag(false);
    setSYNFlag(false);
    setFINFlag(false);
}

TCPService::TCPService(const FWObject *root) : Service(root) 
{
    setInt("src_range_start", 0);
    setInt("src_range_end",   0);
    setInt("dst_range_start", 0);
    setInt("dst_range_end",   0);

    setURGFlag(false);
    setACKFlag(false);
    setPSHFlag(false);
    setRSTFlag(false);
    setSYNFlag(false);
    setFINFlag(false);
}

TCPService::~TCPService() {}

string TCPService::getProtocolName()   {    return "tcp";}
int    TCPService::getProtocolNumber() {    return 6;    }

void TCPService::fromXML(xmlNodePtr root) throw(FWException)
{
    FWObject::fromXML(root);

    const char *n;

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("src_range_start")));
    if(n!=NULL) setStr("src_range_start", n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("src_range_end")));
    if(n!=NULL) setStr("src_range_end", n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("dst_range_start")));
    if(n!=NULL) setStr("dst_range_start", n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("dst_range_end")));
    if(n!=NULL) setStr("dst_range_end", n);



    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("psh_flag")));
    if(n!=NULL) setStr("psh_flag", n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("urg_flag")));
    if(n!=NULL) setStr("urg_flag", n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("syn_flag")));
    if(n!=NULL) setStr("syn_flag", n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("ack_flag")));
    if(n!=NULL) setStr("ack_flag", n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("fin_flag")));
    if(n!=NULL) setStr("fin_flag", n);

    n=FROMXMLCAST(xmlGetProp(root,TOXMLCAST("rst_flag")));
    if(n!=NULL) setStr("rst_flag", n);
}


bool    TCPService::getURGFlag() const { return getBool("urg_flag"); }
bool    TCPService::getACKFlag() const { return getBool("ack_flag"); }
bool    TCPService::getPSHFlag() const { return getBool("psh_flag"); }
bool    TCPService::getRSTFlag() const { return getBool("rst_flag"); }
bool    TCPService::getSYNFlag() const { return getBool("syn_flag"); }
bool    TCPService::getFINFlag() const { return getBool("fin_flag"); }

void    TCPService::setURGFlag(bool f) { setBool("urg_flag", f); }
void    TCPService::setACKFlag(bool f) { setBool("ack_flag", f); }
void    TCPService::setPSHFlag(bool f) { setBool("psh_flag", f); }
void    TCPService::setRSTFlag(bool f) { setBool("rst_flag", f); }
void    TCPService::setSYNFlag(bool f) { setBool("syn_flag", f); }
void    TCPService::setFINFlag(bool f) { setBool("fin_flag", f); }







