/* 

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: Compiler_ops.cc,v 1.16 2002/05/19 04:40:04 vkurland Exp $

*/

#include "Compiler.hh"

#include "fwbuilder/FWObjectDatabase.hh"
#include "fwbuilder/RuleElement.hh"
#include "fwbuilder/Interface.hh"
#include "fwbuilder/Host.hh"
#include "fwbuilder/Network.hh"
#include "fwbuilder/IPService.hh"
#include "fwbuilder/ICMPService.hh"
#include "fwbuilder/TCPService.hh"
#include "fwbuilder/UDPService.hh"
#include "fwbuilder/CustomService.hh"
#include "fwbuilder/Policy.hh"
#include "fwbuilder/Rule.hh"

using namespace fwcompiler;
using namespace libfwbuilder;
using namespace std;

/*
 *  Interface has both address and a netmask, however what we need here is only
 *  its address. If we were to use both, we would get the same behaviour as if
 *  it was network, not just an interface of a single host. 
 */
vector<FWObject*> fwcompiler::_find_obj_intersection(Address *op1, Address *op2)
{
    const IPNetwork n1( op1->getAddress() , 
       	(Interface::cast(op1))?Netmask("255.255.255.255"):op1->getNetmask() );
    const IPNetwork n2( op2->getAddress() , 
	(Interface::cast(op2))?Netmask("255.255.255.255"):op2->getNetmask() );

    vector<IPNetwork> intersection= libfwbuilder::getIntersection(n1,n2);

    vector<FWObject*>  res;

    for (vector<IPNetwork>::iterator i=intersection.begin(); i!=intersection.end(); i++) {
	IPNetwork *n= &(*i);
	if (n->getNetmask()==Netmask("255.255.255.255")) {
	    Host *h=new Host();
	    h->setStr("address",n->getAddress().toString());
	    h->setName("h-"+n->getAddress().toString());
	    op1->getRoot()->addAt(FWObjectDatabase::db->std.HostsId, h);
	    res.push_back(h);
	} else {
	    Network *net=new Network();
	    net->setAddress(n->getAddress());
	    net->setNetmask(n->getNetmask());
	    net->setName("n-"+n->getAddress().toString());
	    op1->getRoot()->addAt(FWObjectDatabase::db->std.NetworksId, net);
	    res.push_back(net);
	}
    }

    return res;

//    throw FWException("can't find intersection of two objects : \
//o1: "+op1->getName()+" ("+op1->getTypeName()+") o2: "+op2->getName()+" ("+op2->getTypeName()+")");
}

bool fwcompiler::_find_portrange_intersection(int rs1,int re1,
					    int rs2,int re2,
					    int &rsr,int &rer)
{
    if (re2<rs1) return false;
    if (rs2<rs1 && re2>rs1 && re2<re1) { rsr=rs1; rer=re2; return true; }
    if (rs2>rs1 && re2<re1) { rsr=rs2; rer=re2; return true; }
    if (rs2>rs1 && rs2<re1 && re2>re1) { rsr=rs2; rer=re1; return true; }
    if (rs2>re1) return false;
    if (rs2<rs1 && re2>re1) { rsr=rs1; rer=re1; return true; }
    if (rs1==rs2 && re1==re2) { rsr=rs1; rer=re1; return true; }
    return false;
}

vector<FWObject*> fwcompiler::_find_srv_intersection(Service *op1, Service *op2)
{
    vector<FWObject*> res;

    if (op1->getTypeName()==op2->getTypeName()) {
	if ( IPService::cast(op1)) {
/* TODO: check ip flags */
	    if (op1->getStr("protocol_num")==op2->getStr("protocol_num")) {
		res.push_back(op1);
	    }
	    return res;
	}
	if ( ICMPService::cast(op1)) {
	    int t1=op1->getInt("type");
	    int t2=op2->getInt("type");

	    if (t1!=-1 && t2!=-1 && t1==t2) { res.push_back(op1); return res; }
	    if (t1!=-1 && t2==-1) { res.push_back(op1); return res; }
	    if (t1==-1 && t2!=-1) { res.push_back(op2); return res; }
	    return res;
	}
	if ( TCPService::cast(op1) || UDPService::cast(op1) ) {
	    int srs1=op1->getInt("src_range_start");
	    int sre1=op1->getInt("src_range_end");
	    int drs1=op1->getInt("dst_range_start");
	    int dre1=op1->getInt("dst_range_end");

	    int srs2=op2->getInt("src_range_start");
	    int sre2=op2->getInt("src_range_end");
	    int drs2=op2->getInt("dst_range_start");
	    int dre2=op2->getInt("dst_range_end");

	    int srsR,sreR,drsR,dreR;

	    if ( ! _find_portrange_intersection(srs1,sre1,srs2,sre2,srsR,sreR) ) return res;
	    if ( ! _find_portrange_intersection(drs1,dre1,drs2,dre2,drsR,dreR) ) return res;
		    
	    FWObject *nserv;
	    if ( TCPService::cast(op1) ) {
		nserv=new TCPService();
		op1->getRoot()->addAt(FWObjectDatabase::db->std.TCPServicesId, nserv);
	    }
	    if ( UDPService::cast(op1) ) {
		nserv=new UDPService();
		op1->getRoot()->addAt(FWObjectDatabase::db->std.UDPServicesId, nserv);
	    }

	    nserv->setName(op1->getName()+"-"+op2->getName());
	    nserv->setInt("src_range_start",srsR);
	    nserv->setInt("src_range_end"  ,sreR);
	    nserv->setInt("dst_range_start",drsR);
	    nserv->setInt("dst_range_end"  ,dreR);

	    res.push_back(nserv);
	    return res;
	}
    }
/*
 * if objects are of the different types, then the only case allowed is
 * when one object is IPService
 */
    if ( ! IPService::cast(op1) && IPService::cast(op2)) res.push_back(op1);
    if ( ! IPService::cast(op2) && IPService::cast(op1)) res.push_back(op2);
    return res;
}



int fwcompiler::cmpRules(const PolicyRule &r1,const PolicyRule &r2)
{
    Address  *src1=r1.getSrc()->getFirst(true);
    Address  *dst1=r1.getDst()->getFirst(true);
    Service  *srv1=r1.getSrv()->getFirst(true);
		           
    Address  *src2=r2.getSrc()->getFirst(true);
    Address  *dst2=r2.getDst()->getFirst(true);
    Service  *srv2=r2.getSrv()->getFirst(true);

//    if (! fwcompiler::intersect( r1, r2 )) return -2;

    if ( (*src1 == *src2) && (*dst1 == *dst2) && (*srv1 == *srv2) ) return  0;
    if ( (*src1 <= *src2) && (*dst1 <= *dst2) && (*srv1 <= *srv2) ) return -1;
    if ( (*src2 <= *src1) && (*dst2 <= *dst1) && (*srv2 <= *srv1) ) return  1;

    return -2;
}

int fwcompiler::cmpRules(const NATRule &r1,const NATRule &r2)
{
    Address  *osrc1=r1.getOSrc()->getFirst(true);
    Address  *odst1=r1.getODst()->getFirst(true);
    Service  *osrv1=r1.getOSrv()->getFirst(true);
			    
    Address  *osrc2=r2.getOSrc()->getFirst(true);
    Address  *odst2=r2.getODst()->getFirst(true);
    Service  *osrv2=r2.getOSrv()->getFirst(true);

    if ( (*osrc1 == *osrc2) && (*odst1 == *odst2) && (*osrv1 == *osrv2) ) return  0;
    if ( (*osrc1 <= *osrc2) && (*odst1 <= *odst2) && (*osrv1 <= *osrv2) ) return -1;
    if ( (*osrc2 <= *osrc1) && (*odst2 <= *odst1) && (*osrv2 <= *osrv1) ) return  1;

    return -2;
}


bool fwcompiler::operator<(const Service &o1,const Service &o2)
{
    if (o1.getTypeName()==o2.getTypeName()) {
	if ( IPService::constcast(&o1)) {
/* TODO: check ip flags */
	    return false;
	}
	if ( ICMPService::constcast(&o1)) {
	    if (o1.getInt("type")!=-1 && o2.getInt("type")==-1) return true;
	    else return false;
	}
	if ( TCPService::constcast(&o1) ) {
            const TCPService *t1=TCPService::constcast(&o1);
            const TCPService *t2=TCPService::constcast(&o2);
            if ( t1->getURGFlag()!=t2->getURGFlag() ||
                 t1->getACKFlag()!=t2->getACKFlag() ||
                 t1->getPSHFlag()!=t2->getPSHFlag() ||
                 t1->getRSTFlag()!=t2->getRSTFlag() ||
                 t1->getSYNFlag()!=t2->getSYNFlag() ||
                 t1->getFINFlag()!=t2->getFINFlag() ) return false;
        }
	if ( TCPService::constcast(&o1) || UDPService::constcast(&o1) ) {
	    int srs1=o1.getInt("src_range_start");
	    int sre1=o1.getInt("src_range_end");
	    int drs1=o1.getInt("dst_range_start");
	    int dre1=o1.getInt("dst_range_end");

	    int srs2=o2.getInt("src_range_start");
	    int sre2=o2.getInt("src_range_end");
	    int drs2=o2.getInt("dst_range_start");
	    int dre2=o2.getInt("dst_range_end");

	    return (srs1>=srs2 && sre1<=sre2 && drs1>=drs2 && dre1<=dre2);
	}
    }
/*
 * if objects are of the different types, then the only case allowed is
 * when one object is IPService
 *
 * TODO: need to take IP flags into account. Typical rule with
 * ip_fragment "shades" everything if we do not check flags here
 *
    if ( ! IPService::constcast(&o1) && IPService::constcast(&o2)) return true;
    if ( ! IPService::constcast(&o2) && IPService::constcast(&o1)) return false;
 */
    if (!o1.isAny() && o2.isAny() ) return true;
    return false;

    throw FWException("trying to compare objects of incompatible types: \
o1: "+o1.getName()+" ("+o1.getTypeName()+") o2: "+o2.getName()+" ("+o2.getTypeName()+")");
}

bool fwcompiler::operator<(const Address &o1,const Address &o2)
{
    IPNetwork n1( o1.getAddress() , 
       	(Interface::constcast(&o1))?Netmask("255.255.255.255"):o1.getNetmask() );
    IPNetwork n2( o2.getAddress() , 
	(Interface::constcast(&o2))?Netmask("255.255.255.255"):o2.getNetmask() );

    vector<IPNetwork> sn = libfwbuilder::getIntersection(n1,n2);

    return (!sn.empty() && sn.front() == n1);
}

/*************************************************************************/




bool fwcompiler::operator==(const Service  &o1,const Service  &o2)
{
    if (o1.getTypeName()==o2.getTypeName()) {
	if ( IPService::constcast(&o1)) {
	    return (o1.getInt("protocol_num")==o2.getInt("protocol_num") &&
                    o1.getStr("fragm")==o2.getStr("fragm") &&
                    o1.getStr("short_fragm")==o2.getStr("short_fragm") &&
                    o1.getStr("lsrr")==o2.getStr("lsrr") &&
                    o1.getStr("ssrr")==o2.getStr("ssrr") &&
                    o1.getStr("rr")==o2.getStr("rr") &&
                    o1.getStr("ts")==o2.getStr("ts") );
	}
	if ( ICMPService::constcast(&o1)) {
	    return (o1.getInt("type")==o2.getInt("type")) &&
		(o1.getInt("code")==o2.getInt("code"));
	}
	if ( TCPService::constcast(&o1) || UDPService::constcast(&o1) ) {
	    int srs1=o1.getInt("src_range_start");
	    int sre1=o1.getInt("src_range_end");
	    int drs1=o1.getInt("dst_range_start");
	    int dre1=o1.getInt("dst_range_end");

	    int srs2=o2.getInt("src_range_start");
	    int sre2=o2.getInt("src_range_end");
	    int drs2=o2.getInt("dst_range_start");
	    int dre2=o2.getInt("dst_range_end");

/* TODO: check tcp flags */
	    return (srs1==srs2 && sre1==sre2 && drs1==drs2 && dre1==dre2);
	}
    }
    return false;
}

bool fwcompiler::operator==(const Address &o1,const Address &o2)
{
    IPNetwork n1( o1.getAddress() , o1.getNetmask() );
    IPNetwork n2( o2.getAddress() , o2.getNetmask() );

    return n1==n2;
}




/*************************************************************************/

bool fwcompiler::operator<=(const Address &o1,const Address &o2)
{
    return (o1 == o2 || o1 < o2);
}

bool fwcompiler::operator<=(const Service  &o1,const Service  &o2)
{
    return (o1 == o2 || o1 < o2);
}


/*************************************************************************/

bool fwcompiler::intersect(const PolicyRule &r1,const PolicyRule &r2)
{
    string act1=r1.getActionAsString();
    string act2=r2.getActionAsString();

    bool res_act;
    res_act=(act1=="CONTINUE" || act2=="CONTINUE" || act1==act2);
    if (res_act==false) return false;

    string iface1=r1.getStr("interface_id");
    string iface2=r2.getStr("interface_id");

    bool res_iface;
    res_iface=(iface1=="" || iface2=="" || iface1==iface2);
    if (res_iface==false) return false;

    vector<FWObject*> v1=_find_obj_intersection( r1.getSrc()->getFirst(true) , r2.getSrc()->getFirst(true) );
    if (v1.empty()) return false;

    vector<FWObject*> v2=_find_obj_intersection( r1.getDst()->getFirst(true) , r2.getDst()->getFirst(true) );
    if (v2.empty()) return false;

    vector<FWObject*> v3=_find_srv_intersection( r1.getSrv()->getFirst(true) , r2.getSrv()->getFirst(true) );
    if (v3.empty()) return false;

    return true;
}


void fwcompiler::getIntersection( const PolicyRule &r1, const PolicyRule &r2, PolicyRule &res)
{
    string act1=r1.getActionAsString();
    string act2=r2.getActionAsString();

/*
 * "CONTINUE" is "broad" action, so chose another one, whatever it is
 */
    if (act1=="CONTINUE") res.setAction(act2);
    if (act2=="CONTINUE") res.setAction(act1);

    string any_id;
    FWObject *any;

    RuleElementSrc *nsrc=res.getSrc();
    nsrc->clearChildren();

    RuleElementDst *ndst=res.getDst();
    ndst->clearChildren();

    RuleElementSrv *nsrv=res.getSrv();
    nsrv->clearChildren();

    if (r1.getStr("interface_id")!=r2.getStr("interface_id")) return ;

    string lbl="'"+r1.getStr("rule_label")+"' & '"+r2.getStr("rule_label")+"'";
    res.setStr("rule_label", lbl );

    vector<FWObject*> v1=_find_obj_intersection( r1.getSrc()->getFirst(true) , r2.getSrc()->getFirst(true) );
    vector<FWObject*> v2=_find_obj_intersection( r1.getDst()->getFirst(true) , r2.getDst()->getFirst(true) );
    vector<FWObject*> v3=_find_srv_intersection( r1.getSrv()->getFirst(true) , r2.getSrv()->getFirst(true) );

    for (vector<FWObject*>::iterator i=v1.begin(); i!=v1.end(); ++i)
	nsrc->addRef(*i);

    for (vector<FWObject*>::iterator i=v2.begin(); i!=v2.end(); ++i)
	ndst->addRef(*i);

    for (vector<FWObject*>::iterator i=v3.begin(); i!=v3.end(); ++i)
	nsrv->addRef(*i);

}
    
