/*

                          Firewall Builder

                 Copyright (C) 2002 Vadim Kurland

  Author:  Vadim Kurland     vadim@vk.crocodile.org

  $Id: PolicyCompiler.cc,v 1.28 2002/06/03 05:15:26 vkurland Exp $

  This program is free software which we release under the GNU General Public
  License. You may redistribute and/or modify this program under the terms
  of that license as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  To get a copy of the GNU General Public License, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/

#ifdef __MINGW32__
#include <assert.h>
#endif

#include "PolicyCompiler.hh"

#include "fwbuilder/AddressRange.hh"
#include "fwbuilder/RuleElement.hh"
#include "fwbuilder/Network.hh"
#include "fwbuilder/IPService.hh"
#include "fwbuilder/ICMPService.hh"
#include "fwbuilder/TCPService.hh"
#include "fwbuilder/UDPService.hh"
#include "fwbuilder/CustomService.hh"
#include "fwbuilder/Policy.hh"
#include "fwbuilder/Rule.hh"
#include "fwbuilder/InterfacePolicy.hh"
#include "fwbuilder/Firewall.hh"
#include "fwbuilder/RuleSet.hh"
#include "fwbuilder/IPAddress.hh"
#include "fwbuilder/Interface.hh"
#include "fwbuilder/FWObjectDatabase.hh"
#include "fwbuilder/XMLTools.hh"
#include "fwbuilder/FWException.hh"
#include "fwbuilder/Group.hh"

#include <iostream>
#include <iomanip>
#include <sstream>

using namespace fwcompiler;
using namespace libfwbuilder;
using namespace std;

int PolicyCompiler::prolog()
{
    Compiler::prolog();

    FWObject  *policy=fw->getFirstByType(Policy::TYPENAME);
    assert(policy);

    combined_ruleset = new Policy();   // combined ruleset (all interface policies and global policy)
    fw->add( combined_ruleset );

    temp_ruleset = new Policy();   // working copy of the policy
    fw->add( temp_ruleset );


/*
 *  build combined policy by collapsing all the rules together.
 *  store ID of the interface in each rule of interface policy.
 *
 *  also calculate global numbers for all rules and store them, too.
 *  These are used to detect rule shading.
 */
    int global_num=0;

    list<FWObject*> l2=fw->getByType(Interface::TYPENAME);
    for (list<FWObject*>::iterator i=l2.begin(); i!=l2.end(); ++i) {
	FWObject *iface=(*i);
	assert(iface);

	FWObject  *ipolicy=iface->getFirstByType(InterfacePolicy::TYPENAME);
	if (ipolicy) { 
	
		list<FWObject*> l3=ipolicy->getByType(PolicyRule::TYPENAME);
		for (list<FWObject*>::iterator j=l3.begin(); j!=l3.end(); ++j) {
		    PolicyRule *r= PolicyRule::cast(*j);
		    if (r->isDisabled()) continue;
		    r->setStr("interface_id",iface->getId());
		    r->setStr("rule_label", createRuleLabel(iface->getName(), r->getPosition()) );
		    r->setInt("rule_number",global_num); global_num++;
		    combined_ruleset->add( r );
		}
	}
    }

	
    list<FWObject*> l3=policy->getByType(PolicyRule::TYPENAME);
    for (list<FWObject*>::iterator j=l3.begin(); j!=l3.end(); ++j) {
	Rule *r= Rule::cast(*j);
	if (r->isDisabled()) continue;
	r->setStr("interface_id","");
	r->setStr("rule_label", createRuleLabel("global", r->getPosition()) );
	r->setInt("rule_number",global_num); global_num++;
	combined_ruleset->add( r );
    }

    initialized=true;
    
    return combined_ruleset->size();
}



bool  PolicyCompiler::ExpandGroups::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    RuleElementSrc *src=rule->getSrc();   assert(src);
    RuleElementDst *dst=rule->getDst();   assert(dst);
    RuleElementSrv *srv=rule->getSrv();   assert(srv);

    compiler->expandGroupsInRuleElement(src);
    compiler->expandGroupsInRuleElement(dst);
    compiler->expandGroupsInRuleElement(srv);
    
    return true;
}



bool  PolicyCompiler::ExpandMultipleAddresses::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    RuleElementSrc *src=rule->getSrc();    assert(src);
    RuleElementDst *dst=rule->getDst();    assert(dst);

    compiler->_expandAddr(rule,src);
    compiler->_expandAddr(rule,dst);

    return true;
}

bool  PolicyCompiler::addressRanges::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    RuleElementSrc *src=rule->getSrc();    assert(src);
    RuleElementDst *dst=rule->getDst();    assert(dst);

    compiler->_expandAddressRanges(rule,src);
    compiler->_expandAddressRanges(rule,dst);

    return true;
}



Rule* PolicyCompiler::getDifference(const PolicyRule &r1, const PolicyRule &r2)
{
    PolicyRule *r = new PolicyRule();
    *r = r1;

    FWObject *nsrc=r->getSrc()->getFirst(true);
    nsrc->clearChildren();

    FWObject *ndst=r->getDst()->getFirst(true);
    ndst->clearChildren();

    Service  *nsrv=r->getSrv()->getFirst(true);
    nsrv->clearChildren();

    if (r1.getStr("interface_id")!=r2.getStr("interface_id")) return r;

/*
    vector<FWObject*> v1=_substract_obj( r1.getSrc() , r2.getSrc() );
    vector<FWObject*> v2=_substract_obj( r1.getDst() , r2.getDst() );
    vector<FWObject*> v3=_substract_srv( r1.getSrv() , r2.getSrv() );

    for (vector<FWObject*>::iterator i=v1.begin(); i!=v1.end(); ++i)
	nsrc->addRef(*i);

    for (vector<FWObject*>::iterator i=v2.begin(); i!=v2.end(); ++i)
	ndst->addRef(*i);

    for (vector<FWObject*>::iterator i=v3.begin(); i!=v3.end(); ++i)
	nsrv->addRef(*i);
*/
    return r;
}



list<FWObject*>::iterator 
PolicyCompiler::find_more_specific_rule(PolicyRule *rule,
					bool check_interface,
					const list<FWObject*>::iterator &start_here,
					const list<FWObject*>::iterator &stop_here,
					PolicyRule **intersection)
{
    list<FWObject*>::iterator  j;
    for (j=start_here ; j!=stop_here; j++) {
	PolicyRule *r = PolicyRule::cast( *j );

	if (! check_interface || (rule->getStr("acl")==r->getStr("acl")) ) {

	    try {
		if (! fwcompiler::intersect( *rule, *r )) continue;

		if (debug>=9) {
		    cerr << "*********  getIntersection: --------------------\n";
		    cerr << debugPrintRule(rule);
		    cerr << debugPrintRule(r);
		    cerr << "------------------------------------------------\n";
		}

		PolicyRule *ir=new PolicyRule();

/* need to place this rule into the tree somewhere so references will
 * get resolved */
		temp_ruleset->add( ir );

/* copy attributes from the current rule we are looking at. Do not change
 * this part of the algorithm as pix compiler relies upon it.
 */
		ir->duplicate(r);
		
		getIntersection(*rule , *r, *ir );

		if (! ir->isEmpty()) {
		    if (debug>=9) {
			cerr << debugPrintRule(ir);
			cerr << "------------------------------------------------\n";
		    }
		    if (intersection!=NULL) *intersection=ir;
		    return j;
		}

	    } catch (FWException ex) {
		cerr << " *** Exception: " << ex.toString() << endl;
	    }  
	}
    }
    return j;
}

bool PolicyCompiler::splitServices::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;


    RuleElementSrv *srv=rule->getSrv();

    if (srv->size()==1) {
	tmp_queue.push_back(rule);
	return true;
    }

    map<int, list<Service*> > services;

 loop:
    for (FWObject::iterator i=srv->begin(); i!=srv->end(); i++) {
	FWObject *o= *i;
	if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));

	Service *s=Service::cast( o );
	assert(s);

	int proto=s->getProtocolNumber();
	services[proto].push_back(s);
    }

    for (map<int, list<Service*> >::iterator i=services.begin(); i!=services.end(); i++) {
	list<Service*> &sl=(*i).second;

	PolicyRule *r= PolicyRule::cast(
	    compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
	compiler->temp_ruleset->add(r);
	r->duplicate(rule);
	RuleElementSrv *nsrv=r->getSrv();
	nsrv->clearChildren();

	for (list<Service*>::iterator j=sl.begin(); j!=sl.end(); j++) {
	    nsrv->addRef( (*j) );
	}

	tmp_queue.push_back(r);
    }
    return true;
}

/*
 * processor splitServices should have been called eariler, so now all
 * services in Srv are of the same type
 */
bool PolicyCompiler::separateTCPWithFlags::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrv *rel= rule->getSrv();

    if (rel->size()==1) {
	tmp_queue.push_back(rule);
	return true;
    }

/* separate TCP services with TCP flags (can't use those in combination
 * with others in multiport 
 */
    list<Service*> services;
    for (FWObject::iterator i=rel->begin(); i!=rel->end(); i++) {
	    
	FWObject *o= *i;
	if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
	    
	TCPService *s=TCPService::cast( o );
	if (s==NULL) continue;

	if (  s->getSYNFlag() ||
	      s->getURGFlag() || 
	      s->getACKFlag() || 
	      s->getPSHFlag() || 
	      s->getRSTFlag() || 
	      s->getFINFlag() ) {

	    PolicyRule *r= PolicyRule::cast(
		compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
	    compiler->temp_ruleset->add(r);
	    r->duplicate(rule);
	    RuleElementSrv *nsrv=r->getSrv();
	    nsrv->clearChildren();
	    nsrv->addRef( s );
	    tmp_queue.push_back(r);
	    services.push_back(s);
	}
    }
    for (list<Service*>::iterator i=services.begin(); i!=services.end(); i++)
	rel->removeRef( (*i) );

    if (!rel->isAny())
	tmp_queue.push_back(rule);

    return true;
}

bool PolicyCompiler::verifyCustomServices::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);

    RuleElementSrv *srv=rule->getSrv();

    for (FWObject::iterator i=srv->begin(); i!=srv->end(); i++) {
	FWObject *o= *i;
	if (FWReference::cast(o)!=NULL) o=compiler->getCachedObj(o->getStr("ref"));
	assert(o!=NULL);
	if (CustomService::isA(o) && 
	    CustomService::cast(o)->getCodeForPlatform(compiler->myPlatformName()).empty())
	    throw FWException("Custom service is not configured for the platform '"+compiler->myPlatformName()+"'. Rule "+rule->getStr("rule_label"));
    }

    return true;
}


bool PolicyCompiler::ConvertToAtomicForAddresses::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrc *src=rule->getSrc();    assert(src);
    RuleElementDst *dst=rule->getDst();    assert(dst);

    for (FWObject::iterator i1=src->begin(); i1!=src->end(); ++i1) {
	for (FWObject::iterator i2=dst->begin(); i2!=dst->end(); ++i2) {

	    PolicyRule *r = PolicyRule::cast(
		compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
	    r->duplicate(rule);
	    compiler->temp_ruleset->add(r);

	    FWObject *s;
	    s=r->getSrc();	assert(s);
	    s->clearChildren();
	    s->add( *i1 );

	    s=r->getDst();	assert(s);
	    s->clearChildren();
	    s->add( *i2 );

	    tmp_queue.push_back(r);
	}
    }
    return true;
}

bool PolicyCompiler::ConvertToAtomicForIntervals::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementInterval *ivl=rule->getWhen();

    if (ivl==NULL || ivl->isAny()) {
        tmp_queue.push_back(rule);
        return true;
    }

    for (FWObject::iterator i1=ivl->begin(); i1!=ivl->end(); ++i1) {

        PolicyRule *r = PolicyRule::cast(
            compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
        r->duplicate(rule);
        compiler->temp_ruleset->add(r);
        
        FWObject *s;

        s=r->getWhen();	assert(s);
        s->clearChildren();
        s->add( *i1 );

        tmp_queue.push_back(r);
    }
    return true;
}

bool  PolicyCompiler::ConvertToAtomic::processNext()
{
    PolicyRule *rule=getNext(); if (rule==NULL) return false;

    RuleElementSrc      *src=rule->getSrc();  assert(src);
    RuleElementDst      *dst=rule->getDst();  assert(dst);
    RuleElementSrv      *srv=rule->getSrv();  assert(srv);

    for (FWObject::iterator i1=src->begin(); i1!=src->end(); ++i1) {
	for (FWObject::iterator i2=dst->begin(); i2!=dst->end(); ++i2) {
	    for (FWObject::iterator i3=srv->begin(); i3!=srv->end(); ++i3) {

                PolicyRule *r = PolicyRule::cast(
                    compiler->dbcopy->create(PolicyRule::TYPENAME,false) );
                r->duplicate(rule);
                compiler->temp_ruleset->add(r);

                FWObject *s;
                s=r->getSrc();	assert(s);
                s->clearChildren();
                s->add( *i1 );

                s=r->getDst();	assert(s);
                s->clearChildren();
                s->add( *i2 );

                s=r->getSrv();	assert(s);
                s->clearChildren();
                s->add( *i3 );

                tmp_queue.push_back(r);
	    }
	}
    }
    return true;
}


std::deque<Rule*>::iterator 
PolicyCompiler::findMoreGeneralRule::find_more_general_rule(PolicyRule *rule,
                                                      bool check_interface,
                                                      const std::deque<Rule*>::iterator &start_here,
                                                      const std::deque<Rule*>::iterator &stop_here)
{
    if (compiler->debug>=9) {
        cerr << "*********  search for more general rule: --------------------\n";
        cerr << compiler->debugPrintRule(rule);
        cerr << endl;
    }

    std::deque<Rule*>::iterator  j;
    for (j=start_here ; j!=stop_here; j++) {
	PolicyRule *r = PolicyRule::cast( *j );

	if (! check_interface || 
	    (rule->getStr("interface_id")==r->getStr("interface_id")) ) {

	    try {
                int cr=fwcompiler::cmpRules( *rule , *r );
                if (cr==-2) {
		    if (compiler->debug>=9) {
			cerr << r->getStr("rule_label") << ": rules do not intersect  \n";
		    }
		    continue;
		}
		if ( cr==-1 ) {
		    if (compiler->debug>=9) {
			cerr << r->getStr("rule_label") << ": FOUND more general rule:\n";
                        cerr << compiler->debugPrintRule(r);
                        cerr << endl;
		    }
		    return j;
		}
		if (compiler->debug>=9) {
                    cerr << r->getStr("rule_label") << ": rule >= r\n";
		}
	    } catch (FWException ex) {
		cerr << " *** Exception: " << ex.toString() << endl;
	    }
	}
    }
    return j;
}


bool PolicyCompiler::DetectShading::processNext()
{
    PolicyRule *rule;
    rule=getNext(); if (rule==NULL) return false;

    tmp_queue.push_back(rule);  // to pass it to the next processor, if any

    std::deque<Rule*>::iterator i=find_more_general_rule(rule,true,
                                                         rules_seen_so_far.begin(),
                                                         rules_seen_so_far.end());
    if (i!=rules_seen_so_far.end()) {
        Rule *r=dynamic_cast<Rule*>(*i);
/*
 * find_more_general finds more general _or_ equivalent rule
 */
        if (r && 
            r->getInt("rule_number") != rule->getInt("rule_number") && 
            ! (*r == *rule) ) 
        {
            throw( FWCompilerRuleShadingException(r,rule) );
        }
    }

    rules_seen_so_far.push_back(rule);

    return true;
}



FWCompilerRuleShadingException::FWCompilerRuleShadingException(Rule *r1,Rule *r2) : 
    FWException("Rule '" + r1->getStr("rule_label") +
		"' shades rule '" + r2->getStr("rule_label") + "'  below it")
{
    rule1=r1;
    rule2=r2;
}

string PolicyCompiler::debugPrintRule(Rule *r)
{
    PolicyRule *rule=PolicyRule::cast(r);

    FWOptions *ruleopt =rule->getOptionsObject();

    RuleElementSrc *srcrel=rule->getSrc();
    RuleElementDst *dstrel=rule->getDst();
    RuleElementSrv *srvrel=rule->getSrv();

    string       iface_id = rule->getStr("interface_id");
    Interface *rule_iface = fw_interfaces[iface_id];

    ostringstream str;

//    str << setw(70) << setfill('-') << "-";

    int no=0;
    FWObject::iterator i1=srcrel->begin();
    FWObject::iterator i2=dstrel->begin(); 
    FWObject::iterator i3=srvrel->begin();
    while ( i1!=srcrel->end() || i2!=dstrel->end() || i3!=srvrel->end() ) {

        str << endl;

        string src=" ";
        string dst=" ";
        string srv=" ";

        if (i1!=srcrel->end()) {
            FWObject *o=*i1;
            if (FWReference::cast(o)!=NULL) o=getCachedObj(o->getStr("ref"));
            src=o->getName();
        }

        if (i2!=dstrel->end()) {
            FWObject *o=*i2;
            if (FWReference::cast(o)!=NULL) o=getCachedObj(o->getStr("ref"));
            dst=o->getName();
        }

        if (i3!=srvrel->end()) {
            FWObject *o=*i3;
            if (FWReference::cast(o)!=NULL) o=getCachedObj(o->getStr("ref"));
            srv=o->getName();
        }

        int w=0;
        if (no==0) {
            str << rule->getStr("rule_label");
            w=rule->getStr("rule_label").length();
        }
        
        str <<  setw(10-w)  << setfill(' ') << " ";

        str <<  setw(18) << setfill(' ') << src.c_str();
        str <<  setw(18) << setfill(' ') << dst.c_str();
        str <<  setw(12) << setfill(' ') << srv.c_str();

        if (no==0) {
            str <<  setw(9)  << setfill(' ') << rule->getActionAsString().c_str();
            str <<  setw(9)  << setfill(' ') << rule->getDirectionAsString().c_str();
            if (rule->getLogging()) str << " LOG";
        } else
            str <<  setw(18)  << setfill(' ') << " ";

        ++no;

        if ( i1!=srcrel->end() ) ++i1;
        if ( i2!=dstrel->end() ) ++i2;
        if ( i3!=srvrel->end() ) ++i3;
    }
    return str.str();
}

